// (C) Copyright 2003: Reece H. Dunn

#ifndef BOOST__PRINT_TYPE__HPP
#define BOOST__PRINT_TYPE__HPP
#  include <boost/io/detail/config.hpp>
#  include <boost/io/format_objects.hpp>

#  include <iostream>
#  include <typeinfo>

#  include <utility>                             // std::pair
#  include <complex>                             // std::complex
#  include <boost/compressed_pair.hpp>           // boost::compressed_pair
#  include <boost/rational.hpp>                  // boost::rational

#  if !defined(BOOST_IOFM__NO_LIB_INTERVAL)
#     include <boost/numeric/interval/interval.hpp> // boost::numeric::interval
#  endif
#  if !defined(BOOST_IOFM__NO_LIB_QUATERNION)
#     include <boost/math/quaternion.hpp>           // boost::math::quaternion
#  endif
#  if !defined(BOOST_IOFM__NO_LIB_OCTONION)
#     include <boost/math/octonion.hpp>             // boost::math::octonion
#  endif

#  include <vector>    // std::vector
#  include <list>      // std::list
#  include <deque>     // std::deque
#  include <map>       // std::map
#  include <set>       // std::set

   namespace boost { namespace io
   {
      template< typename T >
      struct print_type {};

      template< typename T >
      std::ostream & operator<<( std::ostream & os, const print_type< T > & )
      {
         return( os << typeid( T ).name());
      }
#     if !defined(BOOST_IOFM__NO_STD_WSTREAM)
         template< typename T >
         std::wostream & operator<<( std::wostream & os, const print_type< T > & )
         {
#           if defined(__BORLANDC__)
               return( os << typeid( T ).name());
#           else
               const char * str = typeid( T ).name();
               return( os << std::wstring( str, str + std::strlen( str )));
#           endif
         }
#     endif

#     if defined(__GNUC__) // GCC does not have a nice typeid( T ).name(), so lets make one!

         // basic types

         template<> struct print_type< int >
         {
            friend std::ostream & operator<<( std::ostream & os, const print_type )
            {
               return( os << "int" );
            }
         };

         template<> struct print_type< long >
         {
            friend std::ostream & operator<<( std::ostream & os, const print_type )
            {
               return( os << "long" );
            }
         };

         template<> struct print_type< char >
         {
            friend std::ostream & operator<<( std::ostream & os, const print_type )
            {
               return( os << "char" );
            }
         };

         template<> struct print_type< wchar_t >
         {
            friend std::ostream & operator<<( std::ostream & os, const print_type )
            {
               return( os << "wchar_t" );
            }
         };

         template<> struct print_type< float >
         {
            friend std::ostream & operator<<( std::ostream & os, const print_type )
            {
               return( os << "float" );
            }
         };

         template<> struct print_type< double >
         {
            friend std::ostream & operator<<( std::ostream & os, const print_type )
            {
               return( os << "double" );
            }
         };

         template< typename T >
         struct print_type< T * >
         {
            friend std::ostream & operator<<( std::ostream & os, const print_type )
            {
               return ( os << print_type< T >() << " * " );
            }
         };

         // const-volatile

         template< typename T >
         struct print_type< const T >
         {
            friend std::ostream & operator<<( std::ostream & os, const print_type )
            {
               return ( os << "const " << print_type< T >());
            }
         };

         template< typename T >
         struct print_type< volatile T >
         {
            friend std::ostream & operator<<( std::ostream & os, const print_type )
            {
               return ( os << "volatile " << print_type< T >());
            }
         };

         // array

         template< typename T, size_t N >
         struct print_type< T[ N ] >
         {
            friend std::ostream & operator<<( std::ostream & os, const print_type )
            {
               return( os << print_type< T >() << "[ " << N << " ]" );
            }
         };

         // std::pair

         template< typename T1, typename T2 >
         struct print_type< std::pair< T1, T2 > >
         {
            friend std::ostream & operator<<( std::ostream & os, const print_type )
            {
               return( os << "std::pair< " << print_type< T1 >() << ", "
                                           << print_type< T2 >() << " >" );
            }
         };

         // std::complex

         template< typename T >
         struct print_type< std::complex< T > >
         {
            friend std::ostream & operator<<( std::ostream & os, const print_type )
            {
               return( os << "std::complex< " << print_type< T >() << " >" );
            }
         };

         // std::basic_string

         template<>
         struct print_type< std::string >
         {
            friend std::ostream & operator<<( std::ostream & os, const print_type )
            {
               return( os << "std::string" );
            }
         };

#        if !defined(BOOST_IOFM__NO_STD_WSTREAM)
            template<>
            struct print_type< std::wstring >
            {
               friend std::wostream & operator<<( std::wostream & os, const print_type )
               {
                  return( os << "std::wstring" );
               }
            };
#        endif

         // sequential containers

         template< typename T, class Allocator >
         struct print_type< std::vector< T, Allocator > >
         {
            friend std::ostream & operator<<( std::ostream & os, const print_type )
            {
               return( os << "std::vector< " << print_type< T >() << " >" );
            }
         };

         template< typename T, class Allocator >
         struct print_type< std::list< T, Allocator > >
         {
            friend std::ostream & operator<<( std::ostream & os, const print_type )
            {
               return( os << "std::list< " << print_type< T >() << " >" );
            }
         };

         template< typename T, class Allocator >
         struct print_type< std::deque< T, Allocator > >
         {
            friend std::ostream & operator<<( std::ostream & os, const print_type )
            {
               return( os << "std::deque< " << print_type< T >() << " >" );
            }
         };

         // associative containers

         template< typename KeyT, typename T, class Cmp, class Allocator >
         struct print_type< std::map< KeyT, T, Cmp, Allocator > >
         {
            friend std::ostream & operator<<( std::ostream & os, const print_type )
            {
               return( os << "std::map< " << print_type< KeyT >() << ", "
                                          << print_type< T    >() << " >" );
            }
         };

         template< typename KeyT, typename T, class Cmp, class Allocator >
         struct print_type< std::multimap< KeyT, T, Cmp, Allocator > >
         {
            friend std::ostream & operator<<( std::ostream & os, const print_type )
            {
               return( os << "std::multimap< " << print_type< KeyT >() << ", "
                                               << print_type< T    >() << " >" );
            }
         };

         template< typename KeyT, class Cmp, class Allocator >
         struct print_type< std::set< KeyT, Cmp, Allocator > >
         {
            friend std::ostream & operator<<( std::ostream & os, const print_type )
            {
               return( os << "std::set< " << print_type< KeyT >() << " >" );
            }
         };

         template< typename KeyT, class Cmp, class Allocator >
         struct print_type< std::multiset< KeyT, Cmp, Allocator > >
         {
            friend std::ostream & operator<<( std::ostream & os, const print_type )
            {
               return( os << "std::multiset< " << print_type< KeyT >() << " >" );
            }
         };

         // boost::compressed_pair

         template< typename T1, typename T2 >
         struct print_type< boost::compressed_pair< T1, T2 > >
         {
            friend std::ostream & operator<<( std::ostream & os, const print_type )
            {
               return( os << "boost::compressed_pair< " << print_type< T1 >() << ", "
                                                        << print_type< T2 >() << " >" );
            }
         };

         // boost::rational

         template< typename T >
         struct print_type< boost::rational< T > >
         {
            friend std::ostream & operator<<( std::ostream & os, const print_type )
            {
               return( os << "boost::rational< " << print_type< T >() << " >" );
            }
         };

         // boost::numeric::interval

#        if !defined(BOOST_IOFM__NO_LIB_INTERVAL)
            template< typename T, class Traits >
            struct print_type< boost::numeric::interval< T, Traits > >
            {
               friend std::ostream & operator<<( std::ostream & os, const print_type )
               {
                  return( os << "boost::numeric::interval< " << print_type< T >() << " >" );
               }
            };
#        endif

         // boost::numeric::quaternion

#        if !defined(BOOST_IOFM__NO_LIB_QUATERNION)
            template< typename T >
            struct print_type< boost::math::quaternion< T > >
            {
               friend std::ostream & operator<<( std::ostream & os, const print_type )
               {
                  return( os << "boost::math::quaternion< " << print_type< T >() << " >" );
               }
            };
#        endif

         // boost::numeric::octonion

#        if !defined(BOOST_IOFM__NO_LIB_OCTONION)
            template< typename T >
            struct print_type< boost::math::octonion< T > >
            {
               friend std::ostream & operator<<( std::ostream & os, const print_type )
               {
                  return( os << "boost::math::octonion< " << print_type< T >() << " >" );
               }
            };
#        endif

         // outputters

         template<>
         struct print_type< basic_output >
         {
            friend std::ostream & operator<<( std::ostream & os, const print_type )
            {
               return( os << "boost::io::basic_output" );
            }
         };

         template< typename FT, class O1, class O2 >
         struct print_type< pair_output< FT, O1, O2 > >
         {
            friend std::ostream & operator<<( std::ostream & os, const print_type )
            {
               return( os << "boost::io::pair_output< " << print_type< O1 >() << ", "
                                                        << print_type< O2 >() << " >" );
            }
         };

         template< typename FT, class O >
         struct print_type< container_output< FT, O > >
         {
            friend std::ostream & operator<<( std::ostream & os, const print_type )
            {
               return( os << "boost::io::container_output< " << print_type< O >() << " >" );
            }
         };

         template< typename FT, class O >
         struct print_type< array_output< FT, O > >
         {
            friend std::ostream & operator<<( std::ostream & os, const print_type )
            {
               return( os << "boost::io::array_output< " << print_type< O >() << " >" );
            }
         };
#     endif
   }}
#endif
